import hashlib
from datetime import timedelta
from typing import List, Union

import numpy as np
from pandas import DataFrame, concat, to_datetime
from sqlalchemy import Column, and_, func, literal, text
from stakenix import MySQL, PostgreDB
from stakenix.utils.common import merger


def get_traffic_source(df: DataFrame, user_id_column: str = "player_id") -> DataFrame:
    ph = hashlib.md5(b"project").hexdigest()
    projects = [
        "mk5",
        "mk1",
        "mk2",
        "mk3",
        "vipclub",
        "vipclub2",
        "vipclub",
        "lara",
        "vipt",
        "pobeda",
    ]
    regexp = "p[a-z0-9]+p[a-z0-9]+p[a-z0-9]{4}"
    players = concat(
        [
            MySQL(project)
            .query(
                columns=[
                    Column("player_id"),
                    func.regexp_substr(Column("partner_code"), regexp).label(
                        "ref_code"
                    ),
                    Column("partner_subid"),
                ],
                table="players",
                where=Column("player_id").in_(df[user_id_column].unique()),
            )
            .assign(project=project)
            for project in projects
        ]
    )
    unique_projects = players.project.unique()
    campaings = concat(
        [
            MySQL(project).query(
                table="campaigns",
                columns=[
                    Column("id"),
                    Column("name").label("promo_name"),
                    Column("cost_type_label"),
                ],
            )
            for project in projects
        ]
    )

    lp_users = concat(
        [
            MySQL(project)
            .query(table="lp_users", columns=[text("*"), literal(project).label(ph)])
            .rename({"id": "user_id", "login": "Account"}, axis=1)
            for project in projects
        ]
    )
    partners = PostgreDB("customdata").query(
        table="partner_accounts", columns=[text("*")]
    )
    ref_codes = concat(
        [
            MySQL(project).query(
                columns=[
                    func.regexp_substr(Column("ref_code"), regexp).label("ref_code"),
                    Column("campaign_id"),
                    Column("user_id"),
                ],
                table="ref_codes",
            )
            for project in projects
        ]
    )

    merged = merger(
        (df,),
        (players, {"left_on": user_id_column, "right_on": "player_id"}),
        (ref_codes.drop_duplicates(), {"on": "ref_code", "how": "left"}),
        (
            campaings.drop_duplicates(),
            {"left_on": "campaign_id", "right_on": "id", "how": "left"},
        ),
        (lp_users.drop_duplicates(), {"on": "user_id", "how": "left"}),
        (partners.drop_duplicates(), {"on": "Account", "how": "left"}),
    )
    duplicates = merged[merged[user_id_column].duplicated(keep=False)].loc[
        (~merged["ref_code"].isna())
    ]
    to_filter = duplicates.loc[~duplicates[ph].isin(unique_projects)]
    to_filter = to_filter.loc[
        ~(~(to_filter["ref_code"].isna()) & to_filter["Account"].isna())
    ]
    return _set_provider(merged[~merged.index.isin(to_filter.index)]).drop(
        columns=[ph, user_id_column] if user_id_column != "player_id" else [ph]
    )


def _set_provider(df):
    df, col = df.copy(), "source"
    df[col] = df["Provider"].fillna("undefined")
    prv, acc, ref = (df[i] for i in (col, "Account", "ref_code"))
    df[col] = np.where(ref.isnull(), "Direct", prv)
    df[col] = np.where(
        (prv == "undefined") & (acc.isnull() == False), "Partnerka", df[col]
    )
    df[col] = np.where(
        (df[col] == "undefined") & (acc.isnull() == True), "UndefRef", df[col]
    )
    df[col] = np.where(df[col].isin(["Experiments", "Inhouse"]), "Inhouse", df[col])
    df["Provider"] = df["Provider"].fillna(df[col])
    return df


def query_rates(
    dates: List, currencies: Union[List[str], str], delta: timedelta = timedelta(days=1)
):
    dates = [date - delta for date in dates]
    if isinstance(currencies, str):
        currencies = [currencies]
    rates = (
        MySQL(schema="currency_rates")
        .query(
            table="rates",
            columns=[
                Column("date").label("rate_date"),
                Column("symbol").label("currency"),
                Column("rate"),
            ],
            where=and_(
                Column("date").in_(dates),
                Column("symbol").in_(currencies),
            ),
        )
        .replace(to_replace={date: date + delta for date in dates})
    )
    rates["rate_date"] = to_datetime(rates["rate_date"])
    return rates


def _to_usd(
    df: DataFrame,
    values_fields: Union[str, List[str]],
    dates_field: str = "date",
    currencies_field: str = "currency",
) -> DataFrame:
    if df.empty:
        return df
    else:
        try:
            unique_dates = df[~df[dates_field].isna()][dates_field].dt.date.unique()
            df["transaction_date"] = to_datetime(df[dates_field].dt.date)
        except AttributeError:
            dates = df[~df[dates_field].isna()]
            dates[dates_field] = to_datetime(dates[dates_field])
            unique_dates = dates[dates_field].dt.date.unique()
            df["transaction_date"] = to_datetime(to_datetime(df[dates_field]).dt.date)

        currencies = df[currencies_field].unique()
        rates = query_rates(dates=unique_dates, currencies=currencies)
        rated = df.merge(
            rates,
            how="left",
            left_on=["transaction_date", "currency"],
            right_on=["rate_date", "currency"],
        )
        if isinstance(values_fields, list):
            rated[
                [
                    "{column_name}_usd".format(column_name=column)
                    for column in values_fields
                ]
            ] = rated.loc[:, values_fields].div(rated["rate"], axis=0)
            rated = rated.drop(columns=["transaction_date", "rate_date", "rate"])
            return rated
        else:
            rated = rated.assign(
                **{f"{values_fields}_usd": lambda x: x[values_fields] / x["rate"]}
            )
            return rated.drop(columns=["transaction_date", "rate_date", "rate"])


def _to_target_currency(
    df: DataFrame,
    target_currency: str,
    values_fields: Union[str, List[str]],
    dates_field: str = "date",
    currencies_field: str = "currency",
):
    df_in_usd = _to_usd(df, values_fields, dates_field, currencies_field)
    try:
        dates = df_in_usd[dates_field].dt.date.unique()
        df_in_usd["transaction_date"] = to_datetime(df_in_usd[dates_field].dt.date)
    except AttributeError:
        dates = df_in_usd[dates_field].unique()
        df_in_usd["transaction_date"] = to_datetime(
            to_datetime(df_in_usd[dates_field]).dt.date
        )
    rates_in_target_currency = query_rates(dates, target_currency).drop(
        columns=["currency"]
    )
    merged_target_currency_rates = df_in_usd.merge(
        rates_in_target_currency, how="left", left_on="transaction_date", right_on="rate_date"
    )
    if isinstance(values_fields, list):
        merged_target_currency_rates[
            [
                "{column_name}_{target_currency}".format(
                    column_name=column, target_currency=target_currency.lower()
                )
                for column in values_fields
            ]
        ] = merged_target_currency_rates.loc[
            :, [f"{field}_usd" for field in values_fields]
        ].mul(
            merged_target_currency_rates["rate"], axis=0
        )
        merged_target_currency_rates = merged_target_currency_rates.drop(
            columns=["transaction_date", "rate_date", "rate"]
        )
        return merged_target_currency_rates
    else:
        merged_target_currency_rates = merged_target_currency_rates.assign(
            **{
                f"{values_fields}_{target_currency.lower()}": lambda x: x[
                    f"{values_fields}_usd"
                ]
                * x["rate"]
            }
        )
        return merged_target_currency_rates.drop(
            columns=["transaction_date", "rate_date", "rate"]
        )


def get_rates(
    df: DataFrame,
    values_fields: Union[str, List[str]],
    dates_field: str = "date",
    currencies_field: str = "currency",
    target_currency: str = "USD",
) -> DataFrame:
    """Method to convert money sums to usd (or other currency). Currency rate getted on date that was 1 day before transaction

    Args:
        df (DataFrame): Datafra,e with transactions
        values_fields (Union[str, List[str]]): fields that must be converted to usd or other currency
        dates_field (str, optional): column name of df that relates to date of transaction. Defaults to "date".
        currencies_field (str, optional): column name of df that relates to currencies of transactions. Defaults to "currency".
        target_currency (str, optional): currency code in ISO 4217 format. Defaults to "USD".

    Raises:
        KeyError: if values_fields arg is None

    Returns:
        DataFrame: input dataframe with additional column of calculated values in usd or other currency
    """

    if target_currency == "USD":
        result = _to_usd(df, values_fields, dates_field, currencies_field)
        return result
    else:
        result = _to_target_currency(
            df,
            target_currency=target_currency,
            values_fields=values_fields,
            dates_field=dates_field,
            currencies_field=currencies_field,
        )
        return result



from typing import List, Optional, Type, Union

from pandas import DataFrame, read_sql
from pymongo import ASCENDING, DESCENDING
from pymongo.errors import CollectionInvalid
from sqlalchemy import Column, MetaData, Table, select
from sqlalchemy.orm import Query
from sqlalchemy.sql import Alias, Select
from sqlalchemy.sql.elements import BinaryExpression, BooleanClauseList
from sqlalchemy.sql.selectable import CompoundSelect
from stakenix.core.connectors import (
    BigQueryConnector,
    ClickHouseConnector,
    MongoDBDirectConnector,
    MongoDBSSHConnector,
    MsSQLConnector,
    MySQLConnector,
    NoSQLConnector,
    PostgreSQLConnector,
    SQLConnector,
)
from stakenix.utils import bson_to_string, flatten


class Database:

    _cls_connector: Optional[Union[Type[SQLConnector], Type[NoSQLConnector]]]
    _schema: Optional[str] = None
    _database: Optional[str] = None

    def query(self) -> None:
        raise NotImplementedError


class SQLDatabase(Database):
    def __init__(self) -> None:
        self.session = self._cls_connector(
            self._schema, self._database
        )._generate_session()

    def get_table(self, table: str) -> Table:
        return self._get_reflected_table(table)

    def _get_reflected_table(self, table: str) -> Table:
        if not isinstance(table, str):
            raise ValueError("table must be str object")

        return Table(
            table,
            MetaData(),
            autoload_with=self.session.connection(),
            schema=self._schema,
        )

    def build_subquery(self, query: Query) -> Alias:
        return self.build_query(query).subquery()

    def build_query(self, query: Query) -> Query:
        if not isinstance(query, Query):
            raise ValueError("query must be object of Query")

        return query.with_session(self.session)

    def from_query_obj(
        self, query: Union[Query, Select, CompoundSelect], **kwargs
    ) -> DataFrame:
        if isinstance(query, Query):
            query_obj = query.statement
        elif isinstance(query, (Select, Alias, CompoundSelect)):
            query_obj = query
        else:
            raise ValueError("query must be instance of class Query or Select")

        return read_sql(query_obj, self.session.connection(), **kwargs)

    def query(
        self,
        table: str,
        columns: List[Column],
        where: Union[BinaryExpression, BooleanClauseList] = None,
        group_by: Union[Column, List[Column]] = None,
        having: Union[Column, List[Column]] = None,
        order_by: Union[Column, List[Column]] = None,
        limit: int = None,
        in_dict: bool = False,
        **kwargs,
    ) -> DataFrame:

        with self._cls_connector(self._schema, self._database).connection() as sess:
            query = select(
                from_obj=self._get_reflected_table(table),
                columns=columns,
                whereclause=where,
                group_by=group_by,
                having=having,
                order_by=order_by,
                **kwargs,
            ).limit(limit)
            if in_dict:
                cursor = sess.execute(query)
                return [dict(row) for row in cursor]
            else:
                return read_sql(sql=query, con=sess.connection())

    def __del__(self):
        try:
            self.session.close()
        except AttributeError:
            pass


class NoSQLDatabase(Database):
    def __init__(self) -> None:
        super().__init__()

    def query(
        self,
        table: str,
        columns: List[str] = None,
        where: dict = None,
        order_by: Union[List[str], str] = None,
        desc: Union[List[bool], bool] = False,
        limit: int = 0,
        in_dict: bool = False,
    ) -> DataFrame:

        with self._cls_connector(self._schema).connection() as conn:
            if not table in conn.list_collection_names():
                raise CollectionInvalid(
                    f"Collection: {table} doesn't exists in database {conn.name}"
                )
            if order_by is None:
                query = conn[table].find(filter=where, limit=limit, projection=columns)
            else:
                query = (
                    conn[table]
                    .find(filter=where, limit=limit, projection=columns)
                    .sort(
                        [(order_by, DESCENDING if desc else ASCENDING)]
                        if isinstance(order_by, str)
                        else list(
                            zip(
                                order_by,
                                [DESCENDING if level else ASCENDING for level in desc],
                            )
                        )
                    )
                )
            if in_dict:
                return [row for row in query]
            else:
                result = DataFrame(iter(query)).pipe(bson_to_string).pipe(flatten)
                return result


class PostgreDB(SQLDatabase):
    def __init__(self, schema: str = "customdata", database: str = "default") -> None:
        self._cls_connector = PostgreSQLConnector
        self._schema = schema
        self._database = database
        super().__init__()


class MySQL(SQLDatabase):
    def __init__(self, schema: str, database: str = "default") -> None:
        self._cls_connector = MySQLConnector
        self._schema = schema
        self._database = database
        super().__init__()


class MsSQL(SQLDatabase):
    def __init__(self, schema: str = "dbo", database: str = "default") -> None:
        self._cls_connector = MsSQLConnector
        self._schema = schema
        self._database = database
        super().__init__()


class ClickHouse(SQLDatabase):
    def __init__(self, schema: str, database: str = "default") -> None:
        self._cls_connector = ClickHouseConnector
        self._schema = schema
        self._database = database
        super().__init__()


class BigQuery(SQLDatabase):
    def __init__(self, dataset: str, project: str = "default") -> None:
        self._cls_connector = BigQueryConnector
        self._schema = dataset
        self._database = project
        super().__init__()


class MongoDB(NoSQLDatabase):
    def __init__(self, schema: str, via_ssh: bool = True) -> None:
        super().__init__()
        if via_ssh:
            self._cls_connector = MongoDBSSHConnector
        else:
            self._cls_connector = MongoDBDirectConnector

        self._schema = schema



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['stakenix', 'stakenix.core', 'stakenix.utils']

package_data = \
{'': ['*']}

install_requires = \
['PyYAML>=5.4.1,<6.0.0',
 'SQLAlchemy<1.4',
 'clickhouse-driver>=0.2.0,<0.3.0',
 'clickhouse-sqlalchemy>=0.1.5,<0.2.0',
 'google-cloud-bigquery-storage>=2.9.0,<3.0.0',
 'mysql-connector>=2.2.9,<3.0.0',
 'openpyxl>=3.0.7,<4.0.0',
 'pandas>=1.2.3,<2.0.0',
 'psycopg2-binary>=2.8.6,<3.0.0',
 'pyarrow>=5.0.0,<6.0.0',
 'pymongo>=3.11.3,<4.0.0',
 'pyodbc>=4.0.30,<5.0.0',
 'sqlalchemy-bigquery>=1.2.0,<2.0.0',
 'sshtunnel>=0.4.0,<0.5.0']

setup_kwargs = {
    'name': 'stakenix',
    'version': '0.2.2',
    'description': "Package for work with different DB's by a simple interface",
    'long_description': None,
    'author': 'Pavlo Levin',
    'author_email': 'levin@bbq.agency',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7.1,<3.10',
}


setup(**setup_kwargs)



from pathlib import Path
from sys import platform

from yaml import safe_dump

FOLDER_PATH = Path.home().joinpath(".stakenix")
CONFIG_PATH = FOLDER_PATH.joinpath("config.yaml")
GOOGLE_CREDS_PATH = FOLDER_PATH.joinpath("credentials.json")


def generate_config() -> None:
    config = {
        "sql": {
            "mysql": {
                "driver": {"name": "mysql+mysqlconnector", "connection_args": {}},
                "databases": {
                    "default": {
                        "host": "place_for_host",
                        "port": "place_for_port",
                        "username": "place_for_username",
                        "password": "place_for_password",
                    },
                    "lucky_partners": {
                        "host": "place_for_host",
                        "port": "place_for_port",
                        "username": "place_for_username",
                        "password": "place_for_password",
                    },
                    "lucky_partners_backup": {
                        "host": "place_for_host",
                        "port": "place_for_port",
                        "username": "place_for_username",
                        "password": "place_for_password",
                    },
                },
            },
            "postgresql": {
                "driver": {
                    "name": "postgresql",
                    "connection_args": {"options": "-csearch_path={}"},
                },
                "databases": {
                    "default": {
                        "host": "place_for_host",
                        "port": "place_for_port",
                        "username": "place_for_username",
                        "password": "place_for_password",
                        "database_name": "analyst_db",
                    },
                    "alpa": {
                        "host": "place_for_host",
                        "port": "place_for_port",
                        "username": "place_for_username",
                        "password": "place_for_username",
                        "database_name": "rocketplay",
                    },
                    "tableau": {
                        "host": "place_for_host",
                        "port": "place_for_port",
                        "username": "place_for_username",
                        "password": "place_for_password",
                        "database_name": "workgroup",
                    },
                },
            },
            "clickhouse": {
                "driver": {"name": "clickhouse", "connection_args": {}},
                "databases": {
                    "default": {
                        "host": "place_for_host",
                        "port": "place_for_port",
                        "username": "place_for_username",
                        "password": "place_for_username",
                    }
                },
            },
            "mssql": {
                "driver": {
                    "name": "mssql+pymssql" if platform == "linux" else "mssql",
                    "connection_args": {}
                    if platform == "linux"
                    else {"driver": "ODBC Driver 17 for SQL Server"},
                },
                "databases": {
                    "default": {
                        "host": "place_for_host",
                        "port": "place_for_port",
                        "username": "place_for_username",
                        "password": "place_for_password",
                        "database_name": "BPMonline7102CustomerCenterSoftkeyRUS_1453921_0613",
                    }
                },
            },
            "gbq": {
                "driver": {
                    "name": "bigquery",
                    "connection_args": {"credentials_path": f"{GOOGLE_CREDS_PATH}"},
                },
                "databases": {
                    "default": {"database_name": "analytics-316208"},
                    "alaro": {"database_name": "alaro-291913"},
                },
            },
        },
        "nosql": {
            "mongodb": {
                "databases": {
                    "pobeda": {
                        "database_name": "pobeda",
                        "host": "place_for_host",
                        "username": "place_for_username",
                        "password": "place_for_password",
                        "ssh_host": "place_for_ssh_host",
                        "ssh_username": "place_for_username",
                    },
                    "lara": {
                        "database_name": "lotoru",
                        "host": "place_for_host",
                        "username": "place_for_username",
                        "password": "place_for_password",
                        "ssh_host": "place_for_ssh_host",
                        "ssh_username": "place_for_username",
                    },
                    "vipt": {
                        "database_name": "vipt",
                        "host": "place_for_host",
                        "username": "place_for_username",
                        "password": "place_for_password",
                        "ssh_host": "place_for_ssh_host",
                        "ssh_username": "place_for_username",
                    },
                    "mk5": {
                        "database_name": "vulkan",
                        "host": "place_for_host",
                        "username": "place_for_username",
                        "password": "place_for_password",
                        "ssh_host": "place_for_ssh_host",
                        "ssh_username": "place_for_username",
                    },
                    "vipclub": {
                        "database_name": "vipclub",
                        "host": "place_for_host",
                        "username": "place_for_username",
                        "password": "place_for_password",
                        "ssh_host": "place_for_ssh_host",
                        "ssh_username": "place_for_username",
                    },
                }
            }
        },
        "grafana": {
            "url": "place_for_url",
            "headers": {
                "Authorization": "place_for_header",
                "Content-Type": "place_for_header",
                "Accept": "place_for_header",
            },
        },
    }

    if not FOLDER_PATH.exists():
        FOLDER_PATH.mkdir()

    with open(CONFIG_PATH, "w") as file:
        safe_dump(config, file, sort_keys=False)

    print("Config successfully generated")
    print(f"Path to config: {CONFIG_PATH}")


if __name__ == "__main__":

    if CONFIG_PATH.exists():
        print(f"There is already a config. Do you want to rewrite it? [y/n]: ", end="")

    answer = ""
    retries = 3
    while answer not in ["y", "n"]:
        retries -= 1

        answer = input()

        if retries == 0:
            print("\nAttempts exhausted.")
            break

        if answer == "y":
            generate_config()
        elif answer == "n":
            print(f"\nDeclined. Closing module.")
            break
        else:
            print("Wrong input. Possible values [y/n]. Please try again: ", end="")



import json
from abc import ABC, ABCMeta
from contextlib import contextmanager
from os import getenv
from pathlib import Path
from typing import Dict, Optional

from pymongo import MongoClient
from sqlalchemy import create_engine
from sqlalchemy.engine import Connection
from sqlalchemy.engine.url import URL
from sqlalchemy.exc import ProgrammingError
from sshtunnel import SSHTunnelForwarder
from stakenix import Session
from stakenix.config import Config


class SQLConnector(ABC):
    def __init__(self, schema: str, database: str) -> None:
        self.schema: str = schema
        self.database: str = database
        self.connection_args: Optional[Dict] = {}

    def _generate_session(self) -> Session:
        Session.configure(
            bind=create_engine(
                self._create_connection_url(), connect_args=self.connection_args
            )
        )
        return Session()

    @contextmanager
    def connection(self) -> Connection:
        session = self._generate_session()
        try:
            yield session
        except Exception as exc:
            session.rollback()
            raise exc
        finally:
            session.close()

    def _create_connection_url(self) -> str:
        ...


class NoSQLConnector(ABC):
    ssh_pkey: Path = Path.home().joinpath(".ssh").joinpath("id_rsa")


class MySQLConnector(SQLConnector):
    def __init__(self, schema: str, database: str = "default") -> None:
        super().__init__(schema=schema, database=database)

    def _create_connection_url(self) -> str:
        env_db_argument: str = self.database.upper()
        return URL(
            drivername="mysql+mysqlconnector",
            username=getenv(f"MYSQL_{env_db_argument}_USERNAME"),
            password=getenv(f"MYSQL_{env_db_argument}_PASSWORD"),
            host=getenv(f"MYSQL_{env_db_argument}_HOST"),
            port=getenv(f"MYSQL_{env_db_argument}_PORT"),
            database=self.schema,
        )


class PostgreSQLConnector(SQLConnector):
    def __init__(self, schema: str, database: str = "default") -> None:
        super().__init__(schema=schema, database=database)
        self._update_connection_arguments()

    def _update_connection_arguments(self) -> None:
        if self.database == "default":
            self.connection_args: Dict = {"options": f"-csearch_path={self.schema}"}

    def _create_connection_url(self) -> str:
        env_db_argument: str = self.database.upper()
        return URL(
            drivername="postgresql",
            username=getenv(f"PG_{env_db_argument}_USERNAME"),
            password=getenv(f"PG_{env_db_argument}_PASSWORD"),
            host=getenv(f"PG_{env_db_argument}_HOST"),
            port=getenv(f"PG_{env_db_argument}_PORT"),
            database=getenv(f"PG_{env_db_argument}_DB"),
        )


class MsSQLConnector(SQLConnector):
    def __init__(self, schema: str, database: str = "default") -> None:
        super().__init__(schema=schema, database=database)
        self._update_connection_arguments()

    def _update_connection_arguments(self) -> None:
        self.connection_args: Dict = {"driver": "ODBC Driver 17 for SQL Server"}

    def _create_connection_url(self) -> str:
        env_db_argument: str = self.database.upper()
        url = "{driver}://{username}:{password}@{host}:{port}/{schema}"
        return url.format(
            driver="mssql",
            username=getenv(f"MSSQL_{env_db_argument}_USERNAME"),
            password=getenv(f"MSSQL_{env_db_argument}_PASSWORD"),
            host=getenv(f"MSSQL_{env_db_argument}_HOST"),
            port=getenv(f"MSSQL_{env_db_argument}_PORT"),
            schema=getenv(f"MSSQL_{env_db_argument}_DB"),
        )


class ClickHouseConnector(SQLConnector):
    def __init__(self, schema: str, database: str = "default") -> None:
        super().__init__(schema=schema, database=database)

    def _create_connection_url(self) -> str:
        env_db_argument: str = self.database.upper()
        return URL(
            drivername="clickhouse",
            username=getenv(f"CH_{env_db_argument}_USERNAME"),
            password=getenv(f"CH_{env_db_argument}_PASSWORD"),
            host=getenv(f"CH_{env_db_argument}_HOST"),
            port=getenv(f"CH_{env_db_argument}_PORT"),
            database=self.schema,
        )


class BigQueryConnector(SQLConnector):
    def __init__(self, dataset: str, project: str = "") -> None:
        super().__init__(schema=dataset, database=project)
        self.schema: str = dataset
        self.project: str = project.upper()
        self.project_id: Optional[str] = getenv(f"BQ_{self.project}_PROJECT_ID")

    def _generate_credentials(self) -> Path:
        creds_folder: Path = Path.home().joinpath(".google")
        write_creds_path: Path = creds_folder.joinpath(
            f"credentials_{self.project.lower()}.json"
        )

        if not creds_folder.exists():
            creds_folder.mkdir()

        creds: Dict = {
            "type": "service_account",
            "project_id": self.project_id,
            "private_key_id": getenv(f"BQ_{self.project}_PRIVATE_KEY_ID"),
            "private_key": getenv(f"BQ_{self.project}_PRIVATE_KEY").encode("latin1").decode("unicode_escape"), #type: ignore
            "client_email": getenv(f"BQ_{self.project}_CLIENT_EMAIL"),
            "client_id": getenv(f"BQ_{self.project}_CLIENT_ID"),
            "auth_uri": "https://accounts.google.com/o/oauth2/auth",
            "token_uri": "https://oauth2.googleapis.com/token",
            "auth_provider_x509_cert_url": "https://www.googleapis.com/oauth2/v1/certs",
            "client_x509_cert_url": getenv(f"GOOGLE_BQ_{self.project}_CLIENT_CERT_URL"),
        }

        with open(write_creds_path, "w") as file:
            json.dump(creds, file)

        return write_creds_path

    def _generate_session(self) -> Session:
        Session.configure(
            bind=create_engine(
                self._create_connection_url(),
                connect_args=self.connection_args,
                credentials_path=self._generate_credentials(),
            )
        )
        return Session()

    def _create_connection_url(self) -> str:
        return URL(drivername="bigquery", host=self.project_id, database=self.schema)


class MongoDBConnector(NoSQLConnector):
    def __init__(self, schema: str) -> None:
        super().__init__()
        self.schema = schema


class MongoDBSSHConnector(MongoDBConnector):
    def __init__(self, schema: str) -> None:
        super().__init__(schema)

    @contextmanager
    def connection(self) -> MongoClient:
        ssh_params = dict(
            ssh_address_or_host=getenv(f"{self.schema.upper()}_MONGODB_SSH_HOST"),
            ssh_username=getenv(f"{self.schema.upper()}_MONGODB_SSH_USERNAME"),
            ssh_pkey=str(self.ssh_pkey),
            remote_bind_address=(getenv(f"{self.schema.upper()}_MONGODB_HOST"), 27017),
        )
        try:
            ssh = SSHTunnelForwarder(**ssh_params)
            ssh.start()
            mongo_params = dict(
                host=getenv(f"{self.schema.upper()}_MONGODB_HOST"),
                port=ssh.local_bind_port,
                password=getenv(f"{self.schema.upper()}_MONGODB_PASSWORD"),
                username=getenv(f"{self.schema.upper()}_MONGODB_USERNAME"),
            )
            yield MongoClient(**mongo_params)[getenv(f"{self.schema.upper()}_MONGODB_DB")]
        except Exception as exc:
            raise exc
        finally:
            ssh.close()


class MongoDBDirectConnector(MongoDBConnector):
    def __init__(self, schema: str) -> None:
        super().__init__(schema)

    @contextmanager
    def connection(self) -> MongoClient:
        try:
            mongo_params = dict(
                host=getenv(f"{self.schema.upper()}_MONGODB_HOST"),
                port=27017,
                password=getenv(f"{self.schema.upper()}_MONGODB_PASSWORD"),
                username=getenv(f"{self.schema.upper()}_MONGODB_USERNAME"),
            )
            yield MongoClient(**mongo_params)[getenv(f"{self.schema.upper()}_MONGODB_DB")]
        except Exception as exc:
            raise exc
